import torch


class RandomShift(torch.nn.Module):
    def __init__(self, *, padding: int = 4):
        super().__init__()
        self.padding = padding

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        n, c, h, w = x.shape
        device = x.device
        x = torch.nn.functional.pad(
            x, (self.padding, self.padding, self.padding, self.padding)
        )
        eps = 1.0 / (h + 2 * self.padding)
        arange = torch.linspace(-1 + eps, 1 - eps, h, device=device)
        base_grid = torch.stack(torch.meshgrid(arange, arange, indexing="ij"), dim=-1)
        base_grid = base_grid.unsqueeze(0).repeat(n, 1, 1, 1)
        shift = torch.randint(0, 2 * self.padding + 1, (n, 2), device=device)
        shift = shift * 2.0 / (h + 2 * self.padding)
        grid = base_grid + shift.view(n, 1, 1, 2)
        return torch.nn.functional.grid_sample(
            x, grid, padding_mode="zeros", align_corners=False
        )



import torch
import numpy
import time
from ..functions import flush, lambda_returns_quantile


class TrainMixin:
    def __init__(self):
        super().__init__()

    def __initialize_training(self, environment: str, seed: int):
        self.flush_results()
        self.set_precision()
        self.set_seed(seed)

        if getattr(self, "reward_centering"):
            self._average_reward = 0.0

        train_environment, test_environment, action_dimension, observation_shape = (
            self.make_environments(environment=environment, seed=seed)
        )
        self.prepare_network(action_dimension=action_dimension)
        optimizer = self.make_optimizer()
        scaler = torch.amp.GradScaler(enabled=self.device.type == "cuda")

        train_observation, _ = train_environment.reset()
        train_observation = (
            torch.from_numpy(train_observation).to(torch.uint8).to(self.device)
        )

        test_observation, _ = test_environment.reset()
        test_observation = (
            torch.from_numpy(test_observation).to(torch.uint8).to(self.device)
        )
        observation = torch.cat([train_observation, test_observation], dim=0)

        episode_returns = numpy.zeros(self.total_environments, dtype=numpy.float32)

        return (
            train_environment,
            test_environment,
            action_dimension,
            observation_shape,
            optimizer,
            scaler,
            observation,
            episode_returns,
        )

    def __allocate_buffers(
        self, *, observation_shape, action_dimension, is_distributional
    ):
        batch_observations = torch.empty(
            (self.steps_per_update, self.total_environments) + observation_shape,
            dtype=torch.uint8,
            device=self.device,
        )
        batch_actions = torch.empty(
            (self.steps_per_update, self.total_environments),
            dtype=torch.int64,
            device=self.device,
        )
        batch_rewards = torch.empty(
            (self.steps_per_update, self.total_environments),
            dtype=torch.float32,
            device=self.device,
        )
        batch_terminations = torch.empty(
            (self.steps_per_update, self.total_environments),
            dtype=torch.float32,
            device=self.device,
        )

        batch_q = None
        batch_quantiles = None
        if not is_distributional:
            batch_q = torch.empty(
                (self.steps_per_update, self.total_environments, action_dimension),
                dtype=torch.float32,
                device=self.device,
            )
        else:
            batch_quantiles = torch.empty(
                (self.steps_per_update, self.total_environments, self.number_quantiles),
                dtype=torch.float32,
                device=self.device,
            )

        return (
            batch_observations,
            batch_actions,
            batch_rewards,
            batch_terminations,
            batch_q,
            batch_quantiles,
        )

    def __collect_trajectories(
        self,
        *,
        is_distributional,
        frame_count,
        observation,
        train_environment,
        test_environment,
        episode_returns,
        batch_observations,
        batch_actions,
        batch_rewards,
        batch_terminations,
        batch_q,
        batch_quantiles,
    ):
        for step in range(self.steps_per_update):
            train_observation = observation[: self.num_train_environments]
            test_observation = observation[self.num_train_environments :]
            float_train_observations = train_observation.float()
            float_test_observations = test_observation.float()
            epsilon_value = self._network.epsilon.get(
                frame_count,
                self.actual_frames,
            )

            if not is_distributional:
                q_values = self.get_q_values(
                    float_train_observations=float_train_observations,
                    float_test_observations=float_test_observations,
                    gradient=False,
                )
            else:
                q_values, quantiles = self.get_q_and_quantiles(
                    float_train_observations=float_train_observations,
                    float_test_observations=float_test_observations,
                    gradient=False,
                )

            actions_train, actions_test = self.get_actions(
                q_values_train=q_values["train"],
                q_values_test=q_values["test"],
                epsilon_value=epsilon_value,
            )
            actions = numpy.concatenate([actions_train, actions_test], axis=0)

            (
                next_observation_train,
                reward_train,
                termination_train,
                truncation_train,
                info_train,
            ) = train_environment.step(actions_train)

            (
                next_observation_test,
                reward_test,
                termination_test,
                truncation_test,
                info_test,
            ) = test_environment.step(actions_test)

            next_observation = numpy.concatenate(
                [next_observation_train, next_observation_test], axis=0
            )
            rewards = numpy.concatenate([reward_train, reward_test], axis=0)
            terminations = numpy.concatenate(
                [termination_train, termination_test], axis=0
            )
            truncations = numpy.concatenate([truncation_train, truncation_test], axis=0)
            terminations = numpy.logical_or(terminations, truncations)

            reward_train = info_train.get("reward", None)
            reward_test = info_test.get("reward", None)
            if reward_train is not None and reward_test is not None:
                if isinstance(reward_train, numpy.ndarray) and isinstance(
                    reward_test, numpy.ndarray
                ):
                    if reward_train.ndim == 0:
                        rewards = numpy.stack([reward_train, reward_test])
                    else:
                        rewards = numpy.concatenate([reward_train, reward_test], axis=0)
                    episode_returns += rewards

            if getattr(self, "reward_centering"):
                reward_centering_beta = getattr(self, "reward_centering_beta")
                mean_step_reward = numpy.mean(rewards)
                reward_difference = mean_step_reward - self._average_reward
                self._average_reward += reward_centering_beta * reward_difference
                rewards = rewards - self._average_reward

            done_mask = terminations
            if numpy.any(done_mask):
                idx = numpy.nonzero(done_mask)[0]
                scores = episode_returns[done_mask]
                split = idx < self.num_train_environments
                self.results.rewards.train.extend(scores[split].tolist())
                self.results.rewards.test.extend(scores[~split].tolist())
                episode_returns[done_mask] = 0

            batch_observations[step] = observation
            batch_actions[step] = torch.from_numpy(actions).to(self.device)
            batch_rewards[step] = torch.from_numpy(rewards).to(self.device)
            batch_terminations[step] = torch.from_numpy(terminations).to(self.device)

            if not is_distributional:
                batch_q[step] = torch.cat([q_values["train"], q_values["test"]], dim=0)
            else:
                action_idx = (
                    batch_actions[step]
                    .unsqueeze(1)
                    .unsqueeze(2)
                    .expand(-1, self.number_quantiles, -1)
                )
                chosen_quantiles = quantiles.gather(2, action_idx).squeeze(-1)
                batch_quantiles[step] = chosen_quantiles

            observation = (
                torch.from_numpy(next_observation).to(torch.uint8).to(self.device)
            )
            frame_count += self.num_train_environments

        return observation, frame_count

    def __compute_targets(
        self,
        *,
        is_distributional,
        observation,
        batch_q,
        batch_rewards,
        batch_terminations,
        batch_quantiles,
    ):
        if not is_distributional:
            targets = self.get_returns(
                float_observations=observation.float(),
                batch_q=batch_q,
                batch_rewards=batch_rewards,
                batch_terminations=batch_terminations,
            )
        else:
            next_q_values, next_quantiles_all = self.get_q_and_quantiles(
                float_observations=observation.float()
            )
            next_action = next_q_values.argmax(dim=-1, keepdim=True)
            next_action_idx = next_action.unsqueeze(1).expand(
                -1, self.number_quantiles, -1
            )
            next_quantiles = next_quantiles_all.gather(2, next_action_idx).squeeze(-1)

            q_seq_for_lambda = torch.cat(
                [batch_quantiles[1:], next_quantiles.unsqueeze(0)], dim=0
            )

            with torch.no_grad():
                with torch.autocast(device_type=self.device.type, dtype=torch.float16):
                    targets = lambda_returns_quantile(
                        batch_rewards,
                        batch_terminations,
                        q_seq_for_lambda,
                        self.gamma,
                        self.lmbda,
                    )
        return targets

    def __flatten_batches(
        self,
        *,
        is_distributional,
        batch_observations,
        batch_actions,
        targets,
        observation_shape,
    ):
        train_slice = slice(0, self.num_train_environments)
        flattened_observations = batch_observations[
            :, : self.num_train_environments
        ].reshape((-1,) + observation_shape)
        flattened_actions = batch_actions[:, train_slice].reshape(-1)

        if not is_distributional:
            flattened_targets = targets[:, train_slice].reshape(-1)
        else:
            flattened_targets = targets[:, train_slice].reshape(
                -1, self.number_quantiles
            )
        return flattened_observations, flattened_actions, flattened_targets

    def __update_network(
        self,
        *,
        is_distributional,
        optimizer,
        scaler,
        flattened_observations,
        flattened_actions,
        flattened_targets,
    ):
        self._network.train()
        for _ in range(self.epochs):
            indices = torch.randperm(self.batch_size, device=self.device)

            for range_start in range(0, self.batch_size, self.minibatch_size):
                range_end = range_start + self.minibatch_size
                mini_batch_idx = indices[range_start:range_end]
                mini_batch_observations = flattened_observations[mini_batch_idx]
                mini_batch_actions = flattened_actions[mini_batch_idx]
                mini_batch_targets = flattened_targets[mini_batch_idx]

                if not is_distributional:
                    optimizer.zero_grad(set_to_none=True)
                    loss = self.get_loss(
                        mini_batch_observations=mini_batch_observations,
                        mini_batch_targets=mini_batch_targets,
                        mini_batch_actions=mini_batch_actions,
                    )
                    scaler.scale(loss).backward()
                    scaler.unscale_(optimizer)
                    torch.nn.utils.clip_grad_norm_(
                        self._network.parameters(), self.gradient_norm
                    )
                    scaler.step(optimizer)
                    scaler.update()
                    self.results.loss.append(loss.item())
                else:
                    optimizer.quantile_value.zero_grad(set_to_none=True)
                    optimizer.fraction_proposal.zero_grad(set_to_none=True)

                    quantile_loss, fraction_loss = self.get_distributional_loss(
                        mini_batch_observations=mini_batch_observations,
                        mini_batch_actions=mini_batch_actions,
                        mini_batch_targets=mini_batch_targets,
                    )

                    total_loss = quantile_loss + fraction_loss
                    scaler.scale(total_loss).backward()

                    scaler.unscale_(optimizer.quantile_value)
                    scaler.unscale_(optimizer.fraction_proposal)

                    torch.nn.utils.clip_grad_norm_(
                        list(self._network.phi.parameters())
                        + list(self._network.quantile_value.parameters()),
                        self.gradient_norm,
                    )
                    torch.nn.utils.clip_grad_norm_(
                        self._network.fraction_proposal.parameters(), self.gradient_norm
                    )

                    scaler.step(optimizer.quantile_value)
                    scaler.step(optimizer.fraction_proposal)
                    scaler.update()

                    self.results.loss.append(quantile_loss.item())

    def __log_progress(self, *, update, frame_count):
        verbose_window = self.verbose_window
        test_score = (
            0.0
            if len(self.results.rewards.test) < verbose_window
            else numpy.mean(self.results.rewards.test[-verbose_window:])
        )

        if self.verbose and update % self.verbose_interval == 0:
            flush(f"Update {update} | Frames: {frame_count}")
            flush(
                f"Test Score: {test_score:.4f}",
            )

    def train_loop(self, *, environment: str, seed: int):
        frame_count = 0
        is_distributional = self.network not in ["q", "duelling"]

        (
            train_environment,
            test_environment,
            action_dimension,
            observation_shape,
            optimizer,
            scaler,
            observation,
            episode_returns,
        ) = self.__initialize_training(environment=environment, seed=seed)

        (
            batch_observations,
            batch_actions,
            batch_rewards,
            batch_terminations,
            batch_q,
            batch_quantiles,
        ) = self.__allocate_buffers(
            observation_shape=observation_shape,
            action_dimension=action_dimension,
            is_distributional=is_distributional,
        )

        training_start_time = time.time()

        for update in range(1, self.total_updates + 1):
            self._network.eval()

            observation, frame_count = self.__collect_trajectories(
                is_distributional=is_distributional,
                frame_count=frame_count,
                observation=observation,
                train_environment=train_environment,
                test_environment=test_environment,
                episode_returns=episode_returns,
                batch_observations=batch_observations,
                batch_actions=batch_actions,
                batch_rewards=batch_rewards,
                batch_terminations=batch_terminations,
                batch_q=batch_q,
                batch_quantiles=batch_quantiles,
            )

            targets = self.__compute_targets(
                is_distributional=is_distributional,
                observation=observation,
                batch_q=batch_q,
                batch_rewards=batch_rewards,
                batch_terminations=batch_terminations,
                batch_quantiles=batch_quantiles,
            )

            (
                flattened_observations,
                flattened_actions,
                flattened_targets,
            ) = self.__flatten_batches(
                is_distributional=is_distributional,
                batch_observations=batch_observations,
                batch_actions=batch_actions,
                targets=targets,
                observation_shape=observation_shape,
            )

            self.__update_network(
                is_distributional=is_distributional,
                optimizer=optimizer,
                scaler=scaler,
                flattened_observations=flattened_observations,
                flattened_actions=flattened_actions,
                flattened_targets=flattened_targets,
            )

            self.__log_progress(update=update, frame_count=frame_count)

        train_environment.close()
        test_environment.close()
        self.results.duration = time.time() - training_start_time



from .LayerNorm2d import LayerNorm2d
from .Stream import Stream
from .HadamaxBlock import HadamaxBlock
from .EncoderBlock import EncoderBlock
from .FractionProposalStream import FractionProposalStream
from .CosineEmbeddingModule import CosineEmbeddingModule
from .QuantileStream import QuantileStream
from .DuellingQuantileStream import DuellingQuantileStream



from .TrainingResultsMixin import TrainingResultsMixin
from .EnvironmentMixin import EnvironmentMixin
from .ActionsMixin import ActionsMixin
from .EpsilonMixin import EpsilonMixin
from .NetworkMixin import NetworkMixin
from .OptimizerMixin import OptimizerMixin
from .QValueMixin import QValueMixin
from .LossMixin import LossMixin
from .LambdaReturnsMixin import LambdaReturnsMixin
from .TrainMixin import TrainMixin



import torch
from . import LayerNorm2d
from ..constants import ModuleType


class HadamaxBlock(torch.nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: int,
        padding: int,
        stride: int,
        pool_kernel: int,
        pool_stride: int,
        pool_padding: int = 0,
        chi: ModuleType = torch.nn.GELU,
        psi: ModuleType = torch.nn.GELU,
    ):
        super().__init__()
        self.conv_a = torch.nn.Conv2d(
            in_channels,
            out_channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
        )
        self.conv_b = torch.nn.Conv2d(
            in_channels,
            out_channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
        )

        self.norm_a = LayerNorm2d(out_channels)
        self.norm_b = LayerNorm2d(out_channels)

        self.chi = chi()
        self.psi = psi()

        self.pool_kernel = pool_kernel
        self.pool_stride = pool_stride
        self.pool_padding = pool_padding

    def forward(self, x):
        a = self.psi(self.norm_a(self.conv_a(x)))
        b = self.chi(self.norm_b(self.conv_b(x)))

        x = a * b

        return torch.nn.functional.max_pool2d(
            x,
            kernel_size=self.pool_kernel,
            stride=self.pool_stride,
            padding=self.pool_padding,
        )



import torch
from .QuantileStream import QuantileStream


class DuellingQuantileStream(torch.nn.Module):
    def __init__(self, *, action_dimension: int, embedding_dimension: int):
        super().__init__()
        self.value_stream = QuantileStream(
            action_dimension=1,
            embedding_dimension=embedding_dimension,
        )
        self.advantage_stream = QuantileStream(
            action_dimension=action_dimension,
            embedding_dimension=embedding_dimension,
        )

    def forward(self, features: torch.Tensor, tau_hats: torch.Tensor):
        value_quantiles = self.value_stream(features, tau_hats)
        advantage_quantiles = self.advantage_stream(features, tau_hats)
        return (
            value_quantiles
            + advantage_quantiles
            - advantage_quantiles.mean(dim=2, keepdim=True)
        )



import envpool


class EnvironmentMixin:
    def __init__(self):
        super().__init__()

    def make_environments(self, environment: str, seed: int):
        train_environment = envpool.make(
            environment,
            env_type="gymnasium",
            num_envs=self.num_train_environments,
            seed=seed,
            num_threads=self.cpu_count,
            thread_affinity_offset=0,
            noop_max=self.noop,
            reward_clip=self.train_reward_clip,
            episodic_life=self.train_episodic_life,
            frame_skip=self.frame_skip,
            stack_num=self.frame_stack,
        )

        test_environment = envpool.make(
            environment,
            env_type="gymnasium",
            num_envs=self.num_test_environments,
            seed=seed + 1000,
            num_threads=min(self.min_test_cpu_count, self.cpu_count),
            thread_affinity_offset=0,
            noop_max=self.noop,
            reward_clip=self.test_reward_clip,
            episodic_life=self.test_episodic_life,
            frame_skip=self.frame_skip,
            stack_num=self.frame_stack,
        )
        action_dimension = train_environment.action_space.n
        observation_shape = train_environment.observation_space.shape
        return train_environment, test_environment, action_dimension, observation_shape



import torch
from ..functions import mse_loss


class LossMixin:
    def __init__(self):
        super().__init__()

    def get_loss(
        self, mini_batch_observations, mini_batch_actions, mini_batch_targets
    ) -> torch.Tensor:
        with torch.autocast(device_type=self.device.type, dtype=torch.float16):
            q_values = self.get_q_values(
                float_observations=mini_batch_observations.float(),
                gradient=True,
            )
            q_taken = q_values.gather(1, mini_batch_actions.unsqueeze(1)).squeeze()
            return mse_loss(q_taken, mini_batch_targets)

    def get_distributional_loss(
        self, mini_batch_observations, mini_batch_actions, mini_batch_targets
    ):
        with torch.autocast(device_type=self.device.type, dtype=torch.float16):
            mini_batch_observations_float = mini_batch_observations.float()
            features = self._network.phi(mini_batch_observations_float)
            tau, tau_hat, q_probs, entropy = self._network.fraction_proposal(
                features.detach()
            )
            quantiles = self._network.quantile_value(features, tau_hat)

            action_idx = (
                mini_batch_actions.unsqueeze(1)
                .unsqueeze(2)
                .expand(-1, self.number_quantiles, -1)
            )
            current_quantiles = quantiles.gather(2, action_idx).squeeze(-1)

            u = mini_batch_targets.unsqueeze(1) - current_quantiles.unsqueeze(2)
            huber_loss = torch.nn.functional.huber_loss(
                u, torch.zeros_like(u), reduction="none", delta=1.0
            )

            tau_hat_expanded = tau_hat.unsqueeze(2).expand(
                -1, -1, self.number_quantiles
            )
            asym_weights = torch.abs(tau_hat_expanded - (u < 0).float())
            quantile_loss = (asym_weights * huber_loss).sum(dim=1).mean(dim=1).mean()

            with torch.no_grad():
                quantiles_tau = self._network.quantile_value(
                    features.detach(), tau[:, 1:-1]
                )
                action_idx_tau = (
                    mini_batch_actions.unsqueeze(1)
                    .unsqueeze(2)
                    .expand(-1, self.number_quantiles - 1, -1)
                )
                Z_tau = quantiles_tau.gather(2, action_idx_tau).squeeze(-1)

            Z_tau_hat = current_quantiles.detach()
            gradients_tau = 2 * Z_tau - Z_tau_hat[:, :-1] - Z_tau_hat[:, 1:]

            entropy_coeff = getattr(self, "entropy_coef", 0.0)
            fraction_loss = (tau[:, 1:-1] * gradients_tau.detach()).sum(
                dim=1
            ).mean() - entropy_coeff * entropy.mean()

        return quantile_loss, fraction_loss



from types import SimpleNamespace


class OptimizerMixin:
    def __init__(self):
        super().__init__()

    def __build_categorical_optimizers(self):
        image_encoder_parameters = list(self._network.phi.parameters())
        quantile_value_parameters = list(self._network.quantile_value.parameters())
        fraction_proposal_parameters = list(
            self._network.fraction_proposal.parameters()
        )
        ordinary_parameters = image_encoder_parameters + quantile_value_parameters

        quantile_value_optimizer = self.optimizer_instance(
            ordinary_parameters,
            lr=self.lr,
            eps=self.optimizer_epsilon,
        )
        fraction_proposal_optimizer = self.optimizer_instance(
            fraction_proposal_parameters,
            lr=self.fraction_proposal_lr,
            eps=self.optimizer_epsilon,
        )

        return SimpleNamespace(
            **{
                "fraction_proposal": fraction_proposal_optimizer,
                "quantile_value": quantile_value_optimizer,
            }
        )

    def __build_non_categorical_optimizer(self):
        return self.optimizer_instance(
            self._network.parameters(),
            lr=self.lr,
            eps=self.optimizer_epsilon,
            betas=(self.optimizer_first_beta, self.optimizer_second_beta),
            weight_decay=self.optimizer_weight_decay,
        )

    def make_optimizer(self):
        if self.network in ["q", "duelling"]:
            return self.__build_non_categorical_optimizer()
        else:
            return self.__build_categorical_optimizers()



import torch
from ..modules.augmentation import RandomShift
from ..modules.augmentation import ColorIntensity

augmentation_map = {
    "none": torch.nn.Identity(),
    "off": torch.nn.Identity(),
    "shift": RandomShift(),
    "randomshift": RandomShift(),
    "cut": RandomShift(),
    "intensity": ColorIntensity(),
    "color": ColorIntensity(),
    "all": torch.nn.Sequential(RandomShift(), ColorIntensity()),
}



from .FQFNetwork import FQFNetwork
from ..modules import DuellingQuantileStream


class DuellingFQFNetwork(FQFNetwork):
    def __init__(
        self,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.quantile_value = DuellingQuantileStream(
            action_dimension=kwargs["action_dimension"],
            embedding_dimension=kwargs["quantile_embedding_dimension"],
        )



import torch


class EpsilonMixin:
    def __init__(self):
        super().__init__()

    def get_epsilons(self, epsilon_value) -> torch.Tensor:
        training_epsilon_vector = torch.full(
            (self.num_train_environments,),
            epsilon_value,
            device=self.device,
            dtype=torch.float32,
        )
        test_epsilon_vector = torch.zeros(
            (self.num_test_environments,),
            device=self.device,
            dtype=torch.float32,
        )
        return torch.cat([training_epsilon_vector, test_epsilon_vector])



from ..encoders import AlphaEncoder
from ..encoders import BetaEncoder
from ..encoders import GammaEncoder
from ..encoders import DeltaEncoder
from ..encoders import EpsilonEncoder
from ..encoders import EtaEncoder
from ..encoders import ZetaEncoder
from ..encoders import ThetaEncoder
from ..encoders import HadamaxGammaEncoderV1
from ..encoders import HadamaxGammaEncoderV2
from ..encoders import HadamaxNatureDQNEncoder


encoders_map = {
    "alpha": AlphaEncoder,
    "beta": BetaEncoder,
    "gamma": GammaEncoder,
    "delta": DeltaEncoder,
    "epsilon": EpsilonEncoder,
    "eta": EtaEncoder,
    "zeta": ZetaEncoder,
    "theta": ThetaEncoder,
    "hadamax": HadamaxNatureDQNEncoder,
    "dqnhadamax": HadamaxNatureDQNEncoder,
    "pqnhadamax": HadamaxNatureDQNEncoder,
    "hadamaxgammav1": HadamaxGammaEncoderV1,
    "gammahadamaxv1": HadamaxGammaEncoderV1,
    "hadamaxgammav2": HadamaxGammaEncoderV2,
    "gammahadamaxv2": HadamaxGammaEncoderV2,
}



import torch
import numpy
from typing import Tuple
from ..functions import epsilon_greedy_vectorized


class ActionsMixin:
    def __init__(self):
        super().__init__()

    @torch.no_grad()
    def get_actions(
        self,
        q_values_train: torch.Tensor,
        q_values_test: torch.Tensor,
        epsilon_value: float,
    ) -> Tuple[numpy.ndarray, numpy.ndarray]:
        device_type = getattr(self.device, "type", "cpu")
        with torch.autocast(device_type=device_type, dtype=torch.float16):
            if self._network.epsilon_greedy:
                actions_train = epsilon_greedy_vectorized(q_values_train, epsilon_value)
            else:
                actions_train = q_values_train.argmax(dim=-1)

            actions_test = q_values_test.argmax(dim=-1)

        return actions_train.cpu().numpy(), actions_test.cpu().numpy()



import torch
import math
from ..constants import ModuleType
from .Stream import Stream


class CosineEmbeddingModule(torch.nn.Module):
    def __init__(
        self,
        *,
        embedding_dimension: int,
        activation: ModuleType = torch.nn.ReLU,
    ):
        super().__init__()
        self.embedding_dimension = embedding_dimension
        self.cosine_network = Stream(
            input_dimension=embedding_dimension,
            hidden_dimension=embedding_dimension,
            output_dimension=embedding_dimension,
            normalization=False,
        )
        pi_indices = math.pi * torch.arange(1, embedding_dimension + 1).float()
        self.register_buffer("pi_indices", pi_indices)
        self.activation = activation()

    def forward(self, fractions: torch.Tensor):
        cos = torch.cos(fractions.unsqueeze(-1) * self.pi_indices.view(1, 1, -1))
        embeddings = self.cosine_network(cos)
        return self.activation(embeddings)



from baloot import funnel
from pathlib import Path


class TrainingResultsMixin:
    def __init__(self):
        super().__init__()

    def __make_log_filename(self) -> str:
        filename = f"seed-{self.buffer.seed}_"
        filename = f"network-{self.network}_"
        filename = f"environment-{self.buffer.environment}_"
        filename = f"network-{self.network}"
        return f"{filename}.pkl"

    def __build_log_payload(self) -> dict:
        duration = self.results.duration or 0
        return {
            "training_reward": self.results.rewards.train,
            "test_reward": self.results.rewards.test,
            "loss": self.results.loss,
            "duration_seconds": duration,
            "duration_hours": duration / 3600,
        }

    def __create_directory(self, directory_path: str) -> str:
        directory_path = directory_path.replace(".", "/")
        Path(directory_path).mkdir(exist_ok=True, parents=True)
        return directory_path

    def log(self, directory: str = "logs") -> None:
        directory_path = self.__create_directory(directory).strip("/").strip()
        filename = self.__make_log_filename()
        payload = self.__build_log_payload()
        funnel(f"{directory_path}/{filename}", payload)



class LinearEpsilon:
    def __init__(self, span: float = 0.1, minimum: float = 0.001, maximum: float = 1.0):
        self.maximum = maximum
        self.minimum = minimum
        self.span = span

    def get(self, frames, total_frames):
        minimum = self.minimum
        maximum = self.maximum
        decay_duration = total_frames * self.span

        if decay_duration == 0:
            return maximum

        Δ = maximum - minimum
        return max(minimum, maximum - (frames / decay_duration) * Δ)



import torch
from ..constants import ModuleType


class Stream(torch.nn.Module):
    def __init__(
        self,
        *,
        # this is just here to keep the consistency. it doesn't do anything in this block.
        input_dimension: int = 3136,
        hidden_dimension: int = 512,
        output_dimension: int,
        activation: ModuleType = torch.nn.ReLU,
        normalization: bool = True,
    ):
        super().__init__()
        self.first_linear = torch.nn.LazyLinear(hidden_dimension)
        self.second_linear = torch.nn.Linear(hidden_dimension, output_dimension)
        self.normalization_layer = torch.nn.LayerNorm(hidden_dimension)
        self.activation = activation()
        self.normalization = normalization

    def forward(self, features):
        features = self.first_linear(features)
        if self.normalization:
            features = self.normalization_layer(features)
        features = self.activation(features)
        features = self.second_linear(features)
        return features



from ..networks import PQNNetwork, DuellingNetwork, FQFNetwork, DuellingFQFNetwork

networks_map = {
    "q": PQNNetwork,
    "duelling": DuellingNetwork,
    "fqf": FQFNetwork,
    "dfqf": DuellingFQFNetwork,
}
network_map = networks_map



import torch
from ..modules import EncoderBlock
from ..constants import ModuleType


class NatureDQNEncoder(torch.nn.Module):
    def __init__(self, *, activation: ModuleType = torch.nn.ReLU):
        super().__init__()
        self.stream = torch.nn.Sequential(
            EncoderBlock(4, 32, kernel_size=8, stride=4, activation=activation),
            EncoderBlock(32, 64, kernel_size=4, stride=2, activation=activation),
            EncoderBlock(64, 64, kernel_size=3, stride=1, activation=activation),
            torch.nn.Flatten(),
        )

    def forward(self, x):
        return self.stream(x)



import numpy
import torch


def epsilon_greedy_vectorized(q_values, eps):
    if isinstance(eps, (numpy.ndarray, list)):
        eps = torch.tensor(eps, device=q_values.device, dtype=torch.float32)
    elif isinstance(eps, (float, int)):
        eps = torch.tensor([eps], device=q_values.device, dtype=torch.float32)
    elif isinstance(eps, torch.Tensor):
        eps = eps.to(q_values.device)

    if eps.ndim == 1 and eps.shape[0] != q_values.shape[0]:
        eps = eps[0]

    num_envs, action_dim = q_values.shape
    greedy_actions = torch.argmax(q_values, dim=-1)
    random_actions = torch.randint(0, action_dim, (num_envs,), device=q_values.device)

    mask = torch.rand(num_envs, device=q_values.device) < eps
    final_actions = torch.where(mask, random_actions, greedy_actions)
    return final_actions



import warnings

warnings.filterwarnings("ignore")

import torch
import math
import os
from typing import Type
from typing import Literal
from types import SimpleNamespace
from baloot import acceleration_device
from baloot import seed_everything
from .maps import encoders_map
from .maps import acceptable_frames_map
from .maps import augmentation_map
from .mixins import TrainingResultsMixin
from .mixins import EnvironmentMixin
from .mixins import ActionsMixin
from .mixins import EpsilonMixin
from .mixins import NetworkMixin
from .mixins import OptimizerMixin
from .mixins import QValueMixin
from .mixins import LossMixin
from .mixins import LambdaReturnsMixin
from .mixins import TrainMixin


class Aftab(
    TrainingResultsMixin,
    EnvironmentMixin,
    ActionsMixin,
    EpsilonMixin,
    NetworkMixin,
    OptimizerMixin,
    QValueMixin,
    LossMixin,
    LambdaReturnsMixin,
    TrainMixin,
):
    def __init__(
        self,
        *,
        encoder: str | Type[torch.nn.Module] = "gamma",
        network: Literal["q", "duelling", "fqf", "dfqf"] = "q",
        frames: int | Literal["pilot", "full", "ablation"] = "pilot",
        augmentation: Literal["all", "intensity", "shift", "none", "off"] = "shift",
        augmentation_iterations: int = 2,
        frame_skip: int = 4,
        num_minibatches: int = 32,
        epochs: int = 2,
        gamma: float = 0.99,
        lmbda: float = 0.65,
        lr: float = 2.5e-4,
        fraction_proposal_lr: float = 2.5e-9,
        num_train_environments: int = 128,
        num_test_environments: int = 8,
        steps_per_update: int = 32,
        min_test_cpu_count: int = 4,
        noop: int = 30,
        frame_stack: int = 4,
        gradient_norm: float = 10.0,
        verbose: bool = False,
        verbose_interval: int = 10,
        verbose_window: int = 10,
        number_quantiles: int = 32,
        quantile_embedding_dimension: int = 256,
        optimizer_instance: Type[torch.nn.Module] = torch.optim.RAdam,
        optimizer_epsilon: float = 1e-5,
        optimizer_weight_decay: float = 0.0,
        optimizer_first_beta: float = 0.9,
        optimizer_second_beta: float = 0.999,
        should_compile: bool = True,
        train_episodic_life: bool = True,
        test_episodic_life: bool = False,
        train_reward_clip: bool = True,
        test_reward_clip: bool = True,
        reward_centering: bool = True,
        reward_centering_beta: float = 0.01,
    ):
        params = locals()
        params.pop("self")
        self.__initialize_hyperparameters(**params)
        self.__initialize_frames()
        self.__initialize_derived_attributes()
        self.__initialize_constants()
        self.__initialize__encoder()
        self.__initialize_augmentation_pipeline()
        super().__init__()

        self.buffer = SimpleNamespace()

    def __initialize_hyperparameters(self, **hyperparameters):
        for key, value in hyperparameters.items():
            setattr(self, key, value)

    def __initialize_augmentation_pipeline(self):
        try:
            self.augmentation_pipeline = augmentation_map[self.augmentation]
        except KeyError as exc:
            raise ValueError(
                f"Invalid value for `augmentation`: {self.augmentation!r}. "
            ) from exc

    def __initialize_frames(self):
        if not isinstance(self.frames, str):
            return

        try:
            self.frames = acceptable_frames_map[self.frames]
        except KeyError as exc:
            raise ValueError(
                f"Invalid value for `frames`: {self.frames!r}. "
                f"Expected one of {tuple(acceptable_frames_map)}."
            ) from exc

    def __initialize_derived_attributes(self):
        self.total_environments = int(
            self.num_train_environments + self.num_test_environments
        )
        self.batch_size = int(self.num_train_environments * self.steps_per_update)
        self.minibatch_size = int(self.batch_size // self.num_minibatches)
        self.actual_frames = int(self.frames / self.frame_skip)
        self.total_updates = math.ceil(self.actual_frames / self.batch_size)

    def __initialize__encoder(self):
        if not isinstance(self.encoder, str):
            return

        try:
            self.encoder = encoders_map[self.encoder]
        except KeyError as exc:
            raise ValueError(
                f"Unknown encoder key: {self.encoder!r}. "
                f"Expected one of: {tuple(encoders_map.keys())}"
            ) from exc

    def __initialize_constants(self):
        self.device = acceleration_device()
        self.cpu_count = os.cpu_count() or 1

    def set_precision(self):
        torch.set_float32_matmul_precision("high")

    def set_seed(self, seed: int):
        seed_everything(seed)

    def set_buffer(self, key, value):
        setattr(self.buffer, key, value)

    def flush_results(self):
        self.results = SimpleNamespace()
        self.results.rewards = SimpleNamespace()
        self.results.rewards.train = []
        self.results.rewards.test = []
        self.results.loss = []
        self.results.duration = 0.0

    def train(self, *, environment: str, seed: int = 42):
        self.set_buffer("seed", seed)
        self.set_buffer("environment", environment)
        self.train_loop(environment=environment, seed=seed)



import torch


@torch.jit.script
def lambda_returns_quantile(
    rewards,
    terminations,
    next_quantiles,
    gamma: float,
    lmbda: float,
):
    T = next_quantiles.size(0)
    output = torch.zeros_like(next_quantiles)
    not_done_last = (1.0 - terminations[-1]).unsqueeze(-1)
    accumulated = rewards[-1].unsqueeze(-1) + gamma * not_done_last * next_quantiles[-1]
    output[-1] = accumulated
    for t in range(T - 2, -1, -1):
        not_done = (1.0 - terminations[t]).unsqueeze(-1)
        bootstrap = next_quantiles[t]
        mix = (1.0 - lmbda) * bootstrap + lmbda * accumulated
        accumulated = rewards[t].unsqueeze(-1) + gamma * not_done * mix
        output[t] = accumulated
    return output



import torch


class LayerNorm2d(torch.nn.Module):
    def __init__(self, num_features, eps=1e-6):
        super().__init__()
        self.ln = torch.nn.LayerNorm(num_features, eps=eps)

    def forward(self, x):
        x = x.permute(0, 2, 3, 1)
        x = self.ln(x)
        return x.permute(0, 3, 1, 2)



import torch
from ..constants import ModuleType
from ..common import LinearEpsilon
from ..modules.augmentation import RandomShift, ColorIntensity


class BaseNetwork(torch.nn.Module):
    def __init__(
        self, *, action_dimension: int, augmentation: str, encoder: ModuleType
    ):
        super().__init__()

        self.phi = encoder()
        self.epsilon_greedy = True
        self.epsilon = LinearEpsilon()
        self.action_dimension = action_dimension

        if augmentation == "all":
            self.chi = torch.nn.Sequential(RandomShift(), ColorIntensity())
        elif augmentation == "intensity":
            self.chi = ColorIntensity()
        elif augmentation == "shift":
            self.chi = RandomShift()
        elif augmentation in ["none", "off"]:
            self.chi = torch.nn.Identity()
        else:
            raise ValueError(
                f"Augmentation pipeline expected among all, intensity, shift, none, off. Got {augmentation}."
            )

    def no_epsilon_greedy(self):
        self.epsilon_greedy = False

    def normalize_observations(self, x: torch.Tensor) -> torch.Tensor:
        return x / 255.0

    def get_features(self, x: torch.Tensor, augment: bool = True) -> torch.Tensor:
        x = self.normalize_observations(x)

        if augment:
            x = self.chi(x)

        features = self.phi(x)
        return features

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        raise NotImplementedError



import torch
from .Stream import Stream


class FractionProposalStream(torch.nn.Module):
    def __init__(self, *, number_quantiles: int, embedding_dimension: int):
        super().__init__()
        self.number_quantiles = number_quantiles
        self.stream = Stream(
            input_dimension=None,
            hidden_dimension=embedding_dimension,
            output_dimension=number_quantiles,
            normalization=True,
        )

    def forward(self, features):
        q_logits = self.stream(features)
        q_probs = torch.nn.functional.softmax(q_logits, dim=-1)
        q_log_probabilities = torch.log(q_probs.clamp(min=1e-8))
        tau_0 = torch.zeros((features.size(0), 1), device=features.device)
        tau_1_to_N = torch.cumsum(q_probs, dim=-1)
        tau = torch.cat([tau_0, tau_1_to_N], dim=-1)
        tau_hat = (tau[:, :-1] + tau[:, 1:]) / 2.0
        entropy = -torch.sum(
            q_probs * q_log_probabilities,
            dim=-1,
            keepdim=True,
        )
        return tau, tau_hat, q_probs, entropy



from .epsilon_greedy_vectorized import epsilon_greedy_vectorized
from .flush import flush
from .mse_loss import mse_loss
from .lambda_returns import lambda_returns
from .lambda_returns_quantile import lambda_returns_quantile



import torch
from ..maps import networks_map
from ..networks import PQNNetwork
from ..constants import ModuleType


class NetworkMixin:
    def __init__(self):
        super().__init__()

    def __get_dummy_sample(self):
        if not hasattr(self, "frame_stack"):
            raise AttributeError("Expected `frame_stack` to be defined.")

        if not hasattr(self, "device"):
            raise AttributeError("Expected `device` to be defined.")

        batch_size = 1
        picture_size = 84

        return torch.randn(
            batch_size,
            self.frame_stack,
            picture_size,
            picture_size,
            device=self.device,
        )

    @torch.no_grad()
    def __perform_dummy_pass(self):
        if not hasattr(self, "_network"):
            raise AttributeError("Expected `_network` to be defined.")

        dummy_input = self.__get_dummy_sample()
        self._network(dummy_input)

    def __build_pqn_network(self, action_dimension: int, network_instance: ModuleType):
        self._network = network_instance(
            action_dimension=action_dimension,
            encoder=self.encoder,
            augmentation=self.augmentation,
        )

    def __build_categorical_network(
        self, action_dimension: int, network_instance: ModuleType
    ):
        self._network = network_instance(
            action_dimension=action_dimension,
            encoder=self.encoder,
            augmentation=self.augmentation,
            number_quantiles=self.number_quantiles,
            quantile_embedding_dimension=self.quantile_embedding_dimension,
        )

    def __build_network(self, action_dimension: int):
        try:
            network_instance = networks_map[self.network]
            if network_instance == PQNNetwork:
                self.__build_pqn_network(
                    action_dimension=action_dimension, network_instance=network_instance
                )
            else:
                self.__build_categorical_network(
                    action_dimension=action_dimension, network_instance=network_instance
                )
        except Exception as e:
            raise ValueError("Wrong network id detected.", e)

    def __compile_network(self):
        if not getattr(self, "should_compile", False):
            return

        if not hasattr(self, "_network"):
            raise AttributeError(
                "Expected `_network` to be defined before compilation."
            )

        self._network = torch.compile(self._network)

    def __move_network_on_device(self):
        self._network.to(self.device)

    def prepare_network(self, action_dimension: int):
        self.__build_network(action_dimension=action_dimension)
        self.__move_network_on_device()
        self.__perform_dummy_pass()
        self.__compile_network()



import torch


class ColorIntensity(torch.nn.Module):
    def __init__(self, *, scale: float = 0.05):
        super().__init__()
        self.scale = scale

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        noise = 1.0 + self.scale * torch.randn((x.shape[0], 1, 1, 1), device=x.device)
        return x * noise



import torch
from typing import Dict
from .BaseNetwork import BaseNetwork
from ..modules import QuantileStream, FractionProposalStream


class FQFNetwork(BaseNetwork):
    def __init__(
        self,
        quantile_embedding_dimension: int,
        number_quantiles: int,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.fraction_proposal = FractionProposalStream(
            number_quantiles=number_quantiles,
            embedding_dimension=quantile_embedding_dimension,
        )
        self.quantile_value = QuantileStream(
            action_dimension=kwargs["action_dimension"],
            embedding_dimension=quantile_embedding_dimension,
        )

    def get_q(self, x: torch.Tensor) -> torch.Tensor:
        output = self.forward(x)
        quantiles = output["quantiles"]
        q_probs = output["q_probs"]
        return (quantiles * q_probs.unsqueeze(-1)).sum(dim=1)

    def forward(self, x: torch.Tensor) -> Dict:
        features = self.get_features(x)
        taus, tau_hats, q_probs, entropy = self.fraction_proposal(features)
        quantiles = self.quantile_value(features, tau_hats)
        return {
            "features": features,
            "quantiles": quantiles,
            "taus": taus,
            "tau_hats": tau_hats,
            "q_probs": q_probs,
            "entropy": entropy,
        }



import torch
from ..functions import lambda_returns


class LambdaReturnsMixin:
    def __init__(self):
        super().__init__()

    def get_returns(
        self, float_observations, batch_q, batch_rewards, batch_terminations
    ):
        with (
            torch.no_grad(),
            torch.autocast(device_type=self.device.type, dtype=torch.float16),
        ):
            next_q = self._network(float_observations).max(dim=-1).values
            max_q_seq = batch_q.max(dim=-1).values
            q_seq_for_lambda = torch.cat([max_q_seq, next_q.unsqueeze(0)])
            returns = lambda_returns(
                batch_rewards,
                batch_terminations,
                q_seq_for_lambda[1:],
                self.gamma,
                self.lmbda,
            )
        return returns



import torch
from .Stream import Stream
from .CosineEmbeddingModule import CosineEmbeddingModule


class QuantileStream(torch.nn.Module):
    def __init__(
        self,
        *,
        action_dimension: int,
        embedding_dimension: int,
    ):
        super().__init__()
        self.embedding_dimension = embedding_dimension
        self.mu = CosineEmbeddingModule(embedding_dimension=embedding_dimension)
        self.nu = Stream(
            input_dimension=embedding_dimension,
            hidden_dimension=embedding_dimension,
            output_dimension=action_dimension,
        )
        self.xi = torch.nn.Sequential(
            torch.nn.LazyLinear(embedding_dimension),
            torch.nn.ReLU(),
        )

    def forward(self, state_features, fractions):
        psi = self.xi(state_features)
        phi = self.mu(fractions)
        B, N, D = phi.shape
        psi = psi.view(B, 1, D)
        merged = psi * phi
        quantiles = self.nu(merged)
        return quantiles



import torch


class QValueMixin:
    def __init__(self):
        super().__init__()

    def __get_q_value_and_quantiles(
        self, float_observations: torch.Tensor, gradient: bool
    ):
        with (
            torch.set_grad_enabled(gradient),
            torch.autocast(device_type=self.device.type, dtype=torch.float16),
        ):
            features = self._network.phi(float_observations)
            _, tau_hat, q_probs, _ = self._network.fraction_proposal(features)
            quantiles = self._network.quantile_value(features, tau_hat)
            q_values = (q_probs.unsqueeze(-1) * quantiles).sum(dim=1)
            return {"q_values": q_values, "quantiles": quantiles}

    def get_q_values(
        self,
        float_train_observations: torch.Tensor = None,
        float_test_observations: torch.Tensor = None,
        float_observations: torch.Tensor = None,
        gradient: bool = False,
    ):
        with torch.set_grad_enabled(gradient):
            if float_observations is not None:
                return self._network(float_observations)

            test_q_values = self._network(float_test_observations)

            if getattr(self, "augmentation", "none") in ["none", "off"]:
                train_q_values = self._network(float_train_observations)
            else:
                augmentation_iterations = getattr(self, "augmentation_iterations")
                q_values_list = []
                for _ in range(augmentation_iterations):
                    augmented_train_observations = self.augmentation_pipeline(
                        float_train_observations
                    )
                    q_values_list.append(self._network(augmented_train_observations))
                train_q_values = torch.stack(q_values_list).mean(dim=0)

            return {"test": test_q_values, "train": train_q_values}

    def get_q_and_quantiles(
        self,
        float_train_observations: torch.Tensor = None,
        float_test_observations: torch.Tensor = None,
        float_observations: torch.Tensor = None,
        gradient: bool = False,
    ):
        if float_observations is not None:
            res = self.__get_q_value_and_quantiles(float_observations, gradient)
            return res["q_values"], res["quantiles"]

        test_res = self.__get_q_value_and_quantiles(float_test_observations, gradient)
        test_q_values = test_res["q_values"]
        test_quantiles = test_res["quantiles"]

        if getattr(self, "augmentation", "none") in ["none", "off"]:
            train_res = self.__get_q_value_and_quantiles(
                float_train_observations, gradient
            )
            train_q_values = train_res["q_values"]
            train_quantiles = train_res["quantiles"]
        else:
            augmentation_iterations = getattr(self, "augmentation_iterations")
            q_values_list = []
            quantiles_list = []
            for _ in range(augmentation_iterations):
                augmented_train_observations = self.augmentation_pipeline(
                    float_train_observations
                )
                res = self.__get_q_value_and_quantiles(
                    float_observations=augmented_train_observations, gradient=gradient
                )
                q_values_list.append(res["q_values"])
                quantiles_list.append(res["quantiles"])
            train_q_values = torch.stack(q_values_list).mean(dim=0)
            train_quantiles = torch.stack(quantiles_list).mean(dim=0)

        q_values = {"train": train_q_values, "test": test_q_values}
        quantiles = torch.cat([train_quantiles, test_quantiles], dim=0)
        return q_values, quantiles
